"""
caddiepy.
"""

__version__ = "0.2.4"
__author__ = 'Michael Hartung'
__credits__ = """
Prof. Dr. Jan Baumbach, Chair of Computational Systems Biology (https://www.cosy.bio/), University of Hamburg ; 
Dr. Markus List, Head of the Research Group Big Data in BioMedicine (https://biomedical-big-data.de/), Technical University of Munich"""

from .utils.statistics import contingency_table, levenshtein
from .settings import PALETTE_DATASET_COMBINATIONS, PALETTE_PAIRWISE, PALETTE_DRUG_INTERACTION_DATASETS
from .task import Task
from .pipeline import pipeline


import pandas as pd
from caddiepy import settings


def load_csv(file):
    return pd.read_csv(f'{settings.PATH_SOTRAGE}/{file}')

def load_xlsx(file, sheet=0):
    return pd.read_excel(f'{settings.PATH_SOTRAGE}/{file}', sheet_name=sheet)


"""
Handles all communication with the CADDIE API
"""
import requests
import json
import time
from . import settings
import logging

logger = logging.getLogger(__name__)

def get_gene_interaction_datasets():
    query_url = f'{settings.DOMAIN}{settings.Endpoints.GENE_INTERACTION_DATASETS}'
    return requests.get(query_url, headers=settings.HEADERS_JSON, verify=True)

def get_drug_interaction_datasets():
    query_url = f'{settings.DOMAIN}{settings.Endpoints.DRUG_INTERACTION_DATASETS}'
    return requests.get(query_url, headers=settings.HEADERS_JSON, verify=True)

def get_tissues():
    query_url = f'{settings.DOMAIN}{settings.Endpoints.TISSUES}'
    return requests.get(query_url, headers=settings.HEADERS_JSON, verify=True)

def get_expression_cancer_types():
    query_url = f'{settings.DOMAIN}{settings.Endpoints.EXPRESSION_CANCER_TYPES}'
    return requests.get(query_url, headers=settings.HEADERS_JSON, verify=True)

def get_mutation_cancer_types():
    query_url = f'{settings.DOMAIN}{settings.Endpoints.MUTATION_CANCER_TYPES}'
    return requests.get(query_url, headers=settings.HEADERS_JSON, verify=True)

def get_drug_effects():
    query_url = f'{settings.DOMAIN}{settings.Endpoints.DRUG_TARGET_ACTIONS}'
    return requests.get(query_url, headers=settings.HEADERS_JSON, verify=True)

def gene_drug_lookup(search_string, dataset):
    query_url = f'{settings.DOMAIN}{settings.Endpoints.GENE_DRUG_LOOKUP}'
    payload = {
        'text': search_string,
        'dataset': dataset
    }
    return requests.get(query_url, params=payload, headers=settings.HEADERS_JSON, verify=True)

def drug_lookup(search_string, dataset):
    query_url = f'{settings.DOMAIN}{settings.Endpoints.DRUG_LOOKUP}'
    payload = {
        'text': search_string,
        'dataset': dataset
    }
    return requests.get(query_url, params=payload, headers=settings.HEADERS_JSON, verify=True)

def map_gene_id(genes):
    query_url = f'{settings.DOMAIN}{settings.Endpoints.QUERY_NODES}'
    payload = {
        'nodes': genes,
    }
    return requests.post(query_url, data=json.dumps(payload), headers=settings.HEADERS_JSON, verify=True)

def start_task(parameters):
    endpoint = f'{settings.DOMAIN}{settings.Endpoints.TASK}'
    response = requests.post(endpoint, data=parameters, headers=settings.HEADERS_JSON, verify=True)
    return response.json()['token']

def get_task(token, interval=1):
    while True:
        try:
            logger.debug(f'Waiting for task {token}')
            time.sleep(interval)
            
            endpoint_result = f'{settings.DOMAIN}{settings.Endpoints.TASK_RESULT}?token={token}'
            endpoint_task = f'{settings.DOMAIN}{settings.Endpoints.TASK}?token={token}'

            response_task = requests.get(endpoint_task, verify=True).json()
            if response_task['info']['failed']:
                logger.warning(f"task {token} failed due to {response_task['info']['status']}")
                return None
            if not response_task['info']['done']:
                logger.debug(f"task {token} not yet done: {response_task['info']['status']}")
                continue

            response = requests.get(endpoint_result, verify=True)
            # check here if task responded correctly
            if response.status_code != 200:
                logger.warning(f'Failed getting result for {token}')
                continue
            logger.debug(f'Got results for {token}')
            return response.json()
        except:
            logger.info(f'Task {token} is having connection issues, waiting before retrying')
            time.sleep(10)




from enum import Enum
import json
import numpy as np

PATH_SOTRAGE = 'storage'

PALETTE_PAIRWISE = np.array([
    # green
    [105,206,132, 255],
    # grey
    [191, 191, 191, 255]
])/255

PALETTE_DRUG_INTERACTION_DATASETS = np.array([
    # blue
    [46, 134, 193, 255],
    # red
    [231, 76, 60, 255],
    # green
    [40, 180, 99, 255],
    # yellow
    [ 241, 196, 15, 255],
])/255

PALETTE_DATASET_COMBINATIONS = np.array([
    # blue
    [27, 79, 114, 255],
    [33, 97, 140, 255], 
    [46, 134, 193, 255],
    [93, 173, 226, 255],
    [133, 193, 233, 255],
    [199, 220, 240, 255],
    # red
    [ 120, 40, 31, 255],
    [148, 49, 38, 255],
    [ 176, 58, 46, 255 ],
    [231, 76, 60, 255],
    [ 241, 148, 138, 255],
    [ 250, 219, 216, 255],
    # green
    [ 20, 90, 50, 255 ],
    [29, 131, 72, 255],
    [40, 180, 99, 255],
    [ 46, 204, 113, 255],
    [ 125, 206, 160 , 255],
    [  212, 239, 223 , 255],
    # yellow
    [ 125, 102, 8, 255],
    [ 154, 125, 10, 255],
    [ 212, 172, 13, 255],
    [ 241, 196, 15, 255],
    [ 247, 220, 111, 255],
    [ 252, 243, 207, 255],
])/255

class Algorithm(str, Enum):
    TRUSTRANK = 'trustrank'
    DEGREE_CENTRALITY = 'degree_centrality'
    BETWEENNESS_CENTRALITY = 'betweenness_centrality'
    CLOSENESS_CENTRALITY = 'closeness_centrality'
    NETWORK_PROXIMITY = 'network_proximity'
    MUTLI_STEINER = 'multi-steiner'
    KEYPATHWAYMINER = 'keypathwayminer'

class CancerGeneDataset(str, Enum):
    NCG6 = 'NCG6'
    COSMIC = 'COSMIC'
    STRING = 'STRING'
    INTOGEN = 'IntOGen'
    CANCERGENESORG = 'cancer-genes.org'

class GeneInteractionDataset(str, Enum):
    REACTOME = 'REACTOME'
    BIOGRID = 'BioGRID'
    STRING = 'STRING'
    APID = 'APID'
    HTRIdb = 'HTRIdb'
    IID = 'IID'

class DrugInteractionDataset(str, Enum):
    DRUGBANK = 'DrugBank'
    BIOGRID = 'BioGRID'
    CHEMBL = 'ChEMBL'
    DGIDB = 'DGIdb'

class Target(str, Enum):
    DRUG = 'drug'
    DRUGTARGET = 'drug-target'

class DrugEffect(str, Enum):
    NONE = ''
    INHIBITOR = 'inhibitor'
    NOT_INHIBITOR = 'not_inhibitor'
    ACTIVATOR = 'activator'
    NOT_ACTIVATOR = 'not_activator'

class TaskParameters:
    def __init__(self) -> None:
        self.seeds = []
        self.cancer_dataset = CancerGeneDataset('NCG6')
        self.gene_interaction_datasets = [GeneInteractionDataset('IID')]
        self.drug_interaction_datasets = [DrugInteractionDataset('DrugBank')]
        self.cancer_types = []
        self.include_nutraceutical_drugs = True
        self.only_atc_l_drugs = False
        self.filter_paths = True
        self.mutation_cancer_type = None
        self.expression_cancer_type = None
        self.drug_target_action = None
        self.include_indirect_drugs = True
        self.include_non_approved_drugs = True
        self.ignore_non_seed_baits = True
        self.hub_penalty = 0
        self.result_size = 20
        self.available_drugs = None
        self.drug_target_action = DrugEffect('')
        self.cancer_type_names = []
        # multi steiner
        self.num_trees = 5
        self.tolerance = 10
        # trustrank
        self.damping_factor = 0.85
        # keypathwayminer
        self.k = 20


class TaskConfig:
    def __init__(self) -> None:
        # default parameter
        self.algorithm = Algorithm('trustrank')
        self.target = Target('drug')
        self.parameters = TaskParameters()

    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, 
            sort_keys=True, indent=4)
    
DOMAIN = 'https://exbio.wzw.tum.de/caddie/api/'

HEADERS_JSON = {
                'Content-type':'application/json', 
                'Accept':'application/json'
            }

class Endpoints:
    QUERY_NODES = 'query_nodes/'  # POST cancer_dataset, nodes, cancer_types
    TASK = 'task/'  # POST (start task), GET (get task information)
    TASK_RESULT = 'task_result/' # GET (get task result)
    DRUG_LOOKUP = 'drug_interaction_lookup/'
    GENE_DRUG_LOOKUP = 'gene_drug_interaction_lookup/'
    GENE_INTERACTION_DATASETS = 'interaction_gene_datasets/'
    DRUG_INTERACTION_DATASETS = 'interaction_drug_datasets/'
    TISSUES = 'tissues/'
    EXPRESSION_CANCER_TYPES = 'expression_cancer_types/'
    MUTATION_CANCER_TYPES = 'mutation_cancer_types/'
    DRUG_TARGET_ACTIONS = 'drug_target_actions/'



import urllib.request as request
from pathlib import Path
import zipfile
from caddiepy import settings


def download_file_ftp(filename, filepath):
    Path(settings.PATH_SOTRAGE).mkdir(parents=True, exist_ok=True)

    request.urlretrieve(filepath, f'{settings.PATH_SOTRAGE}/{filename}')


def unzip(file):
    with zipfile.ZipFile(f'{settings.PATH_SOTRAGE}/{file}', 'r') as zip_ref:
        zip_ref.extractall(f'{settings.PATH_SOTRAGE}/{file[:-4]}')


from .task import Task
from . import api


def pipeline(gene_list):

    # verify that genes exist in CADDIE
    caddie_gene_list = api.map_gene_id(gene_list).json()['genes']

    task = Task('drug-target', 'multi-steiner', [gene['graphId'] for gene in caddie_gene_list])
    # customize parameters
    task.set_parameter('result_size', 20)
    task.set_parameter('gene_interaction_datasets', ['BioGRID'])
    # start task
    task.run()
    # get result when task is finished, if task is still running wait until it is finished
    result_drug_targets = task.get_result()

    task = Task('drug', 'trustrank', result_drug_targets['network']['nodes'])
    # customize parameters
    task.set_parameter('result_size', 20)
    task.set_parameter('include_indirect_drugs', True)
    task.set_parameter('gene_interaction_datasets', ['IID'])
    task.set_parameter('drug_interaction_datasets', ['DrugBank'])
    # start task
    task.run()
    # get result when task is finished, if task is still running wait until it is finished
    result = task.get_result()

    return result


from setuptools import setup

setup(
    name='caddiepy',
    version='0.2.4',
    author='Maiykol',
    author_email='michael.hartung@uni-hamburg.de',
    packages=['caddiepy', 'caddiepy.examples', 'caddiepy.file_utils', 'caddiepy.utils'],
    scripts=[],
    url='http://pypi.python.org/pypi/caddiepy/',
    license='LICENSE',
    description='The python package to the Cancer Driver Drug Interaction Explorer (CADDIE)',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        "numpy>=1.19.5",
        "pandas>=1.1.5",
        "requests==2.26.0",
        "scipy>=1.5.4",
        "seaborn==0.11.2",
        "openpyxl==3.0.7"
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        # 'License :: OSI Approved :: BSD License',  
        # 'Operating System :: POSIX :: Linux',        
        # 'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 2.7',
        # 'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



"""
Handles tasks and the corresponding parameter
"""
from . import settings
import caddiepy.api as api


class Task:
    def __init__(self, target, algorithm, seeds):
        self.config = settings.TaskConfig()
        self._set_target(target)
        self._set_algorithm(algorithm)
        self.set_parameter('seeds', seeds)

    def _set_algorithm(self, algorithm):
        self.config.algorithm = settings.Algorithm(algorithm)
    
    def _set_target(self, target):
        self.config.target = settings.Target(target)

    def _start_task(self):
        self.token = api.start_task(self.config.to_json())
    
    def _get_task(self):
        self.result = api.get_task(self.token)

    def get_result(self):
        return self.result

    def set_parameter(self, parameter, value):
        if parameter == 'gene_interaction_datasets':
            setattr(self.config.parameters, parameter, [settings.GeneInteractionDataset(x).value for x in value])
        elif parameter == 'drug_interaction_datasets':
            setattr(self.config.parameters, parameter, [settings.DrugInteractionDataset(x).value for x in value])
        elif parameter == 'cancer_gene_dataset':
            setattr(self.config.parameters, parameter, settings.CancerGeneDataset(value).value)
        elif parameter == 'drug_target_action':
            setattr(self.config.parameters, parameter, settings.DrugEffect(value).value)
        else:
            setattr(self.config.parameters, parameter, value)

    def run(self):
        self._start_task()
        self._get_task()


import numpy as np

ALPHABET = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']

def contingency_table(pred, test, cutoff=10):
    # calculates contingency table comparing a list of predictions "pred" to list of expected elements "test"
    # lists need to be of same length and contain the same elements
    # cut lists based of cutoff
    test_cutoff = test[:cutoff]
    tp = 0
    fp = 0
    tn = 0
    fn = 0
    for i, x in enumerate(pred):
        if i < cutoff:
            if x in test_cutoff:
                tp += 1
            else:
                fp += 1
        else:  # i > cutoff
            if x in test_cutoff:
                fn += 1
            else:
                tn += 1
    return [[tp, fp], [fn, tn]]


def levenshtein(seq1, seq2):
    size_x = len(seq1) + 1
    size_y = len(seq2) + 1
    matrix = [[0. for x in range(size_y)] for y in range(size_x)]
    for x in range(size_x):
        matrix[x][0] = x
    for y in range(size_y):
        matrix[0][y] = y

    for x in range(1, size_x):
        for y in range(1, size_y):
            if seq1[x-1] == seq2[y-1]:
                matrix[x][y] = min(
                    matrix[x-1][y] + 1,
                    matrix[x-1][y-1],
                    matrix[x][y-1] + 1
                )
            else:
                matrix[x][y] = min(
                    matrix[x-1][y] + 1,
                    matrix[x-1][y-1] + 1,
                    matrix[x][y-1] + 1
                )
    return matrix[size_x - 1][size_y - 1]